from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union

from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.language_models.base import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables.base import Runnable, RunnableLambda
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from langchain_ollama import ChatOllama
from langchain_ollama.embeddings import OllamaEmbeddings
from pydantic.main import BaseModel

from llm_sdk.model_providers.base import AIModel


class OllamaLargeLanguageModel(AIModel):
    """
    Model class for Ollama large language model.
    """

    _model: Optional[BaseChatModel] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        self._credential = credential
        self._model_name = model_name
        self._model = ChatOllama(
            base_url=credential.get("base_url", ""),
            model=model_name,
        )

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs
    ) -> Any:
        return self._model.invoke(input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs
    ) -> Any:
        return await self._model.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        self._model.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self._model.astream(input_dict, config=config, **kwargs):
            yield chunk

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        self._tools = tools
        if self._model is not None:
            self._model = self._model.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        def wrapper(input_data: LanguageModelInput, config: Optional[RunnableConfig] = None):
            structured_llm = self._model.with_structured_output(schema=schema, **kwargs)

            return structured_llm.invoke(input_data, config=config)

        return RunnableLambda(wrapper)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_base": self._credential.get("base_url", ""),
            "openai_api_type": "ollama",
            "openai_api_deployment": self._model_name,
        }


class OllamaEmbeddingModel(AIModel):

    _model: Optional[Embeddings] = None

    def __init__(self, mode_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()
        self._model_name = mode_name
        self._credential = credential

        self._model = OllamaEmbeddings(
            model=mode_name,
            base_url=credential.get("base_url", ""),
        )

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self._model.embed_documents(texts)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self._model.aembed_documents(texts)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self._model.embed_query(text)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self._model.aembed_query(text)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_base": self._credential.get("base_url", ""),
            "openai_api_type": "ollama",
            "openai_api_deployment": self._model_name,
        }



import time
from abc import ABC, abstractmethod
from typing import Optional

from pydantic import BaseModel


class RerankDocument(BaseModel):
    """
    Model class for rerank document.
    """

    index: int
    text: str
    score: float


class RerankResult(BaseModel):
    """
    Model class for rerank result.
    """

    docs: list[RerankDocument]


class RerankModel(ABC):
    """
    Base class for rerank models.
    """

    @abstractmethod
    def rerank(
        self,
        query: str,
        docs: list[str],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
    ) -> RerankResult:
        """
        Invoke rerank model

        :param model: model name
        :param credentials: model credentials
        :param query: search query
        :param docs: docs for reranking
        :param score_threshold: score threshold
        :param top_n: top n
        :param user: unique user id
        :return: rerank result
        """
        self.started_at = time.perf_counter()

        try:
            return self._invoke(query, docs, score_threshold, top_n)
        except Exception as e:
            raise self._transform_invoke_error(e)



import base64
import hashlib
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes


class AESCipher:
    def __init__(self, key_string):
        """
        initialize AES cipher
        :param key_string: key string, will be converted to 32 bytes key
        """
        # use SHA-256 to convert arbitrary length string to 32 bytes key
        self.key = hashlib.sha256(key_string.encode()).digest()
        self.backend = default_backend()

    def encrypt(self, plaintext):
        """
        encrypt string
        :param plaintext: string to encrypt
        :return: base64 encoded encrypted string
        """
        # generate random IV (initialization vector)
        iv = os.urandom(16)

        # convert string to bytes
        plaintext_bytes = plaintext.encode("utf-8")

        # create AES-CBC encryptor
        encryptor = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=self.backend).encryptor()

        # ensure plaintext length is a multiple of 16 (AES block size)
        padder = self._create_padder()
        padded_plaintext = padder.update(plaintext_bytes) + padder.finalize()

        # encrypt
        ciphertext = encryptor.update(padded_plaintext) + encryptor.finalize()

        # combine IV and ciphertext, then base64 encode
        encrypted_data = base64.b64encode(iv + ciphertext).decode("utf-8")

        return encrypted_data

    def decrypt(self, encrypted_data):
        """
        decrypt string
        :param encrypted_data: base64 encoded encrypted string
        :return: decrypted string
        """
        try:
            # base64 decode
            encrypted_bytes: bytes = base64.b64decode(encrypted_data)

            # extract IV (first 16 bytes)
            iv = encrypted_bytes[:16]
            ciphertext = encrypted_bytes[16:]

            # create AES-CBC decryptor
            decryptor = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=self.backend).decryptor()

            # decrypt
            padded_plaintext = decryptor.update(ciphertext) + decryptor.finalize()

            # remove padding
            unpadder = self._create_unpadder()
            plaintext_bytes = unpadder.update(padded_plaintext) + unpadder.finalize()

            # convert to string
            plaintext = plaintext_bytes.decode("utf-8")

            return plaintext
        except Exception as e:
            raise ValueError(f"decrypt failed: {str(e)}")

    def _create_padder(self):
        """create PKCS7 padder"""
        from cryptography.hazmat.primitives.padding import PKCS7

        return PKCS7(algorithms.AES.block_size).padder()

    def _create_unpadder(self):
        """create PKCS7 unpadder"""
        from cryptography.hazmat.primitives.padding import PKCS7

        return PKCS7(algorithms.AES.block_size).unpadder()


if __name__ == "__main__":
    # this is a Cipher test
    cipher = AESCipher("kb_ai_agent")

    # text for test
    # plaintext = "This is plain text"
    # encrypted = cipher.encrypt(plaintext)
    # print(f"加密后: {encrypted}")

    # decrypt
    decrypted = cipher.decrypt(
        "PX784ZDszIUME3nd9G4I+S+PlojNEfDbV4OwGX2O9+0="
    )
    print(f"decrypted: {decrypted}")



import logging
from typing import Dict, Optional

from llm_sdk.model_providers.azure_openai import (
    AzureOpenAIEmbeddingModel,
    AzureOpenAILargeLanguageModel,
)
from llm_sdk.model_providers.base import AIModel, ConfigType
from llm_sdk.model_providers.deepseek import DeepseekLargeLanguageModel
from llm_sdk.model_providers.ollama import (
    OllamaEmbeddingModel,
    OllamaLargeLanguageModel,
)
from llm_sdk.model_providers.openai_api_compatible import (
    OAIAPICompatEmbeddingModel,
    OAIAPICompatLargeLanguageModel,
    OAICompatRerankModel,
)
from llm_sdk.model_providers.tongyi import (
    GTERerankModel,
    TongyiLargeLanguageModel,
    TongyiTextEmbeddingModel,
)
from llm_sdk.model_providers.xinference import (
    XinferenceAILargeLanguageModel,
    XinferenceRerankModel,
    XinferenceTextEmbeddingModel,
)

logger = logging.getLogger(__name__)


class ModelProviderFactory:

    @staticmethod
    def build_provider_model(
        model_provider: str,
        model_name: str,
        config_type: ConfigType = ConfigType.CHAT,
        crendential: Optional[Dict] = None,
        app_id: Optional[str] = None,
    ) -> Optional[AIModel]:
        provider_map = {
            "azure_openai": {
                ConfigType.CHAT: AzureOpenAILargeLanguageModel,
                ConfigType.EMBEDDING: AzureOpenAIEmbeddingModel,
            },
            "xinference": {
                ConfigType.CHAT: XinferenceAILargeLanguageModel,
                ConfigType.EMBEDDING: XinferenceTextEmbeddingModel,
                ConfigType.RERANK: XinferenceRerankModel,
            },
            "tongyi": {
                ConfigType.CHAT: TongyiLargeLanguageModel,
                ConfigType.EMBEDDING: TongyiTextEmbeddingModel,
                ConfigType.RERANK: GTERerankModel,
            },
            "openai_api_compatible": {
                ConfigType.CHAT: OAIAPICompatLargeLanguageModel,
                ConfigType.EMBEDDING: OAIAPICompatEmbeddingModel,
                ConfigType.RERANK: OAICompatRerankModel,
            },
            "ollama": {
                ConfigType.CHAT: OllamaLargeLanguageModel,
                ConfigType.EMBEDDING: OllamaEmbeddingModel,
            },
            "deepseek": {
                ConfigType.CHAT: DeepseekLargeLanguageModel,
            },
        }
        provider = provider_map.get(model_provider)
        if provider:
            model_class = provider.get(config_type)
            if model_class:
                return model_class(model_name, crendential, app_id)
            else:
                raise ValueError(f"Model class not found for {model_provider} and {config_type}")
        else:
            raise ValueError(f"Provider not found for {model_provider}")



from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union

from langchain_community.embeddings.xinference import XinferenceEmbeddings
from langchain_community.llms.xinference import Xinference
from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.language_models.base import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables.base import Runnable, RunnableLambda
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from langchain_openai.chat_models.base import ChatOpenAI
from pydantic import SecretStr
from pydantic.main import BaseModel
from xinference_client.client.restful.restful_client import (
    RESTfulRerankModelHandle,
)

from llm_sdk.model_providers.aes import AESCipher
from llm_sdk.model_providers.base import AIModel
from llm_sdk.model_providers.errors import InvokeServerUnavailableError
from llm_sdk.model_providers.rerank import RerankDocument, RerankResult

import logging

logger = logging.getLogger(__name__)


class XinferenceAILargeLanguageModel(AIModel):

    _model: Optional[BaseChatModel] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("server_url"):
                credential["server_url"] = aes_client.decrypt(credential["server_url"])
            else:
                raise ValueError("server_url is required")
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        api_base = credential.get("server_url")
        if api_base:
            api_base = api_base.removesuffix("/")
            if not api_base.endswith("/v1"):
                api_base = api_base + "/v1"
        self._model = ChatOpenAI(
            model=model_name,
            api_key=credential.get("api_key") or SecretStr("sk-xxxxxx"),
            base_url=api_base,
            temperature=credential.get("temperature", 0.7),
            timeout=credential.get("timeout", 60),
            max_retries=3,
        )
        self._credential = credential
        self._model_name = model_name

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return self._model.invoke(input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return await self._model.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        self._model.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self._model.astream(input_dict, config=config, **kwargs):
            yield chunk

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        self._tools = tools
        if self._model is not None:
            self._model = self._model.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        def wrapper(input_data: LanguageModelInput, config: Optional[RunnableConfig] = None):
            structured_llm = self._model.with_structured_output(schema=schema, **kwargs)

            return structured_llm.invoke(input_data, config=config)

        return RunnableLambda(wrapper)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("api_key") or SecretStr("sk-xxxxxx"),
            "openai_api_base": self._credential.get("server_url"),
            "openai_api_version": self._credential.get("model_uid", ""),
            "openai_api_type": "xinference",
            "openai_api_deployment": self._model_name,
        }


class XinferenceTextEmbeddingModel(AIModel):

    _model: Optional[Embeddings] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("server_url"):
                credential["server_url"] = aes_client.decrypt(credential["server_url"])
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = XinferenceEmbeddings(
            server_url=credential.get("server_url"),
            model_uid=credential.get("model_uid", ""),
        )
        self._credential = credential
        self._model_name = model_name

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self._model.embed_documents(texts)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self._model.aembed_documents(texts)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self._model.embed_query(text)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self._model.aembed_query(text)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("api_key") or SecretStr("sk-xxxxxx"),
            "openai_api_base": self._credential.get("server_url"),
            "openai_api_version": self._credential.get("model_uid", ""),
            "openai_api_type": "xinference",
            "openai_api_deployment": self._model_name,
        }


class XinferenceRerankModel(AIModel):

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("server_url"):
                credential["server_url"] = aes_client.decrypt(credential["server_url"])
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._credential = credential
        self._model_name = model_name

    def rerank(
        self,
        query: str,
        docs: list[str],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
    ) -> Any:
        """
        Invoke rerank model

        :param model: model name
        :param credentials: model credentials
        :param query: search query
        :param docs: docs for reranking
        :param score_threshold: score threshold
        :param top_n: top n
        :param user: unique user id
        :return: rerank result
        """
        if len(docs) == 0:
            return RerankResult(docs=[])

        server_url = self._credential.get("server_url", "")
        if not server_url:
            raise InvokeServerUnavailableError("server_url is required")
        model_uid = self._credential.get("model_uid", "")
        api_key = self._credential.get("api_key", "")
        server_url = server_url.removesuffix("/")
        auth_headers = {"Authorization": f"Bearer {api_key}"} if api_key else {}

        params = {"documents": docs, "query": query, "top_n": top_n, "return_documents": True}
        try:
            handle = RESTfulRerankModelHandle(model_uid, server_url, auth_headers)
            response = handle.rerank(**params)
        except RuntimeError as e:
            if "rerank hasn't support extra parameter" not in str(e):
                raise InvokeServerUnavailableError(str(e))

            # compatible xinference server between v0.10.1 - v0.12.1, not support 'return_len'
            handle = RESTfulRerankModelHandleWithoutExtraParameter(model_uid, server_url, auth_headers)
            response = handle.rerank(**params)

        rerank_documents = []
        for idx, result in enumerate(response["results"]):
            # format document
            index = result["index"]
            page_content = (
                result["document"] if isinstance(result["document"], str) else result["document"]["text"]
            )
            rerank_document = RerankDocument(
                index=index,
                text=page_content,
                score=result["relevance_score"],
            )

            # score threshold check
            if score_threshold is None or result["relevance_score"] >= score_threshold:
                rerank_documents.append(rerank_document)

        return RerankResult(docs=rerank_documents)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("api_key") or SecretStr("sk-xxxxxx"),
            "openai_api_base": self._credential.get("server_url"),
            "openai_api_version": self._credential.get("model_uid", ""),
            "openai_api_type": "xinference",
            "openai_api_deployment": self._model_name,
        }


class RESTfulRerankModelHandleWithoutExtraParameter(RESTfulRerankModelHandle):
    def rerank(
        self,
        documents: list[str],
        query: str,
        top_n: Optional[int] = None,
        max_chunks_per_doc: Optional[int] = None,
        return_documents: Optional[bool] = None,
        **kwargs,
    ):
        url = f"{self._base_url}/v1/rerank"
        request_body = {
            "model": self._model_uid,
            "documents": documents,
            "query": query,
            "top_n": top_n,
            "max_chunks_per_doc": max_chunks_per_doc,
            "return_documents": return_documents,
        }

        import requests

        response = requests.post(url, json=request_body, headers=self.auth_headers)
        if response.status_code != 200:
            raise InvokeServerUnavailableError(
                f"Failed to rerank documents, detail: {response.json()['detail']}"
            )
        response_data = response.json()
        return response_data



from typing import Optional


class InvokeError(ValueError):
    """Base class for all LLM exceptions."""

    description: Optional[str] = None

    def __init__(self, description: Optional[str] = None) -> None:
        self.description = description

    def __str__(self):
        return self.description or self.__class__.__name__


class InvokeServerUnavailableError(InvokeError):
    """Raised when the Invoke returns server unavailable error."""

    description = "Server Unavailable Error"


class CredentialsValidateFailedError(ValueError):
    """
    Credentials validate failed error
    """

    pass


class InternalServerError(Exception):
    pass


class InvalidAuthenticationError(Exception):
    pass


class InvalidAPIKeyError(Exception):
    pass


class RateLimitReachedError(Exception):
    pass


class InsufficientAccountBalanceError(Exception):
    pass


class BadRequestError(Exception):
    pass



from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union

from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.language_models import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables import Runnable, RunnableLambda
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools import BaseTool
from langchain_openai.chat_models.azure import AzureChatOpenAI
from langchain_openai.embeddings.azure import AzureOpenAIEmbeddings
from pydantic.main import BaseModel

from llm_sdk.model_providers.aes import AESCipher
from llm_sdk.model_providers.base import AIModel

import logging

logger = logging.getLogger(__name__)


class AzureOpenAILargeLanguageModel(AIModel):

    _model: Optional[BaseChatModel] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("openai_api_key"):
                credential["openai_api_key"] = aes_client.decrypt(credential["openai_api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = AzureChatOpenAI(
            model=credential.get("base_model_name", ""),
            azure_endpoint=credential.get("openai_api_base", ""),
            api_key=credential.get("openai_api_key", ""),
            api_version=credential.get("openai_api_version", ""),
            temperature=credential.get("temperature", 0.7),
            timeout=credential.get("timeout", 60),
            max_retries=3,
        )
        self._credential = credential

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return self._model.invoke(input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return await self._model.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        self._model.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self._model.astream(input_dict, config=config, **kwargs):
            yield chunk

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        self._tools = tools
        if self._model is not None:
            self._model = self._model.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        def wrapper(input_data: LanguageModelInput, config: Optional[RunnableConfig] = None):
            structured_llm = self._model.with_structured_output(schema=schema, **kwargs)

            return structured_llm.invoke(input_data, config=config)

        return RunnableLambda(wrapper)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("openai_api_key", ""),
            "openai_api_base": self._credential.get("openai_api_base", ""),
            "openai_api_version": self._credential.get("openai_api_version", ""),
            "openai_api_type": "azure_openai",
            "openai_api_deployment": self._credential.get("base_model_name", ""),
            "timeout": self._credential.get("timeout", 60),
            "max_retries": self._credential.get("max_retries", 1),
            "temperature": self._credential.get("temperature", 0.7),
        }


class AzureOpenAIEmbeddingModel(AIModel):

    _model: Optional[Embeddings] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("openai_api_key"):
                credential["openai_api_key"] = aes_client.decrypt(credential["openai_api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = AzureOpenAIEmbeddings(
            model=credential.get("base_model_name", ""),
            azure_endpoint=credential.get("openai_api_base", ""),
            api_version=credential.get("openai_api_version", ""),
            api_key=credential.get("openai_api_key", ""),
            timeout=credential.get("timeout", 60),
        )
        self._credential = credential
        self._model_name = model_name

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self._model.embed_documents(texts)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self._model.aembed_documents(texts)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self._model.embed_query(text)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self._model.aembed_query(text)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("openai_api_key", ""),
            "openai_api_base": self._credential.get("openai_api_base", ""),
            "openai_api_version": self._credential.get("openai_api_version", ""),
            "openai_api_type": "azure_openai",
            "openai_api_deployment": self._credential.get("base_model_name", ""),
        }



from enum import Enum
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Iterator,
    List,
    Optional,
    Type,
    Union,
)

from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables import Runnable
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from pydantic.main import BaseModel

from llm_sdk.model_providers.rerank import RerankModel


class ConfigType(Enum):
    CHAT = "CHAT"
    EMBEDDING = "EMBEDDING"
    RERANK = "RERANK"


class AIModel(Runnable, Embeddings, RerankModel):

    def __init__(self):
        super().__init__()

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs
    ) -> Any:
        raise NotImplementedError("Chat is not supported for this model")

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs
    ) -> Any:
        raise NotImplementedError("Chat is not supported for this model")

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        raise NotImplementedError("Chat is not supported for this model")

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        raise NotImplementedError("Chat is not supported for this model")

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        raise NotImplementedError("Embedding is not supported for this model")

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        raise NotImplementedError("Embedding is not supported for this model")

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        raise NotImplementedError("Embedding is not supported for this model")

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        raise NotImplementedError("Embedding is not supported for this model")

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        raise NotImplementedError("Chat is not supported for this model")

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        raise NotImplementedError("Chat is not supported for this model")

    def rerank(
        self,
        query: str,
        docs: list[str],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
    ):
        raise NotImplementedError("Rerank is not supported for this model")

    def unified_credential(credential: Dict) -> Dict:
        raise NotImplementedError("Provider is not supported for this model")



from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union

from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables import Runnable
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from pydantic.main import BaseModel

from llm_sdk.model_provider_factory import ModelProviderFactory
from llm_sdk.model_providers.base import AIModel, ConfigType
import requests


class ModelInstance(Runnable):
    """
    A ModelInstance class that can be used to create a LLM instance based on the given tool name.
    """

    model_insntance: Optional[AIModel] = None

    def __init__(
        self,
        config_type: ConfigType = ConfigType.CHAT,
        model_params: Optional[Dict] = None,
        app_id: Optional[str] = "chatbot",
    ):
        self._tools: Optional[List[BaseTool]] = None
        self.model_insntance = ModelProviderFactory().build_provider_model(
            model_provider=model_params.get("provider"),
            model_name=model_params.get("model"),
            config_type=config_type,
            crendential=model_params.get("credentials", {}),
            app_id=app_id,
        )

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return self.model_insntance.invoke(input_dict=input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return await self.model_insntance.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        return self.model_insntance.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self.model_insntance.astream(input_dict, config=config, **kwargs):
            yield chunk

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self.model_insntance.embed_documents(texts, config=config)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self.model_insntance.aembed_documents(texts, config=config)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self.model_insntance.embed_query(text, config=config)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self.model_insntance.aembed_query(text, config=config)

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "ModelInstance":
        self._tools = tools
        if self.model_insntance is not None:
            self.model_insntance = self.model_insntance.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        return self.model_insntance.with_structured_output(schema=schema, **kwargs)

    def rerank(
        self,
        input: Union[List[BaseMessage], List[str], str],
        documents: List[str],
        top_n: Optional[int] = None,
        score_threshold: Optional[float] = None,
    ):
        return self.model_insntance.rerank(
            query=input, docs=documents, top_n=top_n, score_threshold=score_threshold
        )

    def unified_credential(self) -> Dict:
        return self.model_insntance.unified_credential()


class ModelInstanceFactory:

    @staticmethod
    def get_model_instance(
        model_managment_url: str,
        config_type: ConfigType = ConfigType.CHAT,
        model_params: Optional[Dict] = None,
        app_id: Optional[str] = None,
        app_name: Optional[str] = None,
        token: Optional[str] = None,
    ) -> ModelInstance:
        """_summary_

        Args:
            model_managment_url (str): model management url
            config_type (ConfigType, optional): Defaults to ConfigType.CHAT.
            app_id (Optional[str], optional): Defaults to None.

        Returns:
            ModelInstance: _description_
        """
        if model_params:
            return ModelInstance(
                config_type=config_type,
                model_params=model_params,
                app_id=app_id,
            )

        request_url = f"{model_managment_url.rstrip('/')}/v1.0/system/provider-model"
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
        }
        if config_type == ConfigType.CHAT:
            model_type = "llm"
        elif config_type == ConfigType.EMBEDDING:
            model_type = "text-embedding"
        elif config_type == ConfigType.RERANK:
            model_type = "rerank"
        params = {
            "agent_name": app_name,
            "model_type": model_type,
        }

        response = requests.get(request_url, headers=headers, params=params)
        response.raise_for_status()

        model_params = response.json()

        model_instance = ModelInstance(
            config_type=config_type,
            model_params=model_params,
            app_id=app_id,
        )

        return model_instance



import json
from json import dumps
from threading import Lock
from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union
from urllib.parse import urljoin

import httpx
import numpy as np
import requests
from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.language_models.base import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables.base import Runnable, RunnableLambda
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from langchain_openai.chat_models.base import ChatOpenAI
from pydantic.main import BaseModel
from requests import post
from yarl import URL

from llm_sdk.model_providers.aes import AESCipher
from llm_sdk.model_providers.base import AIModel
from llm_sdk.model_providers.errors import (
    CredentialsValidateFailedError,
    InvokeServerUnavailableError,
)
from llm_sdk.model_providers.rerank import RerankDocument, RerankResult
import logging

logger = logging.getLogger(__name__)


_tokenizer: Any = None
_lock = Lock()


class OAIAPICompatLargeLanguageModel(AIModel):
    """
    Model class for Openai compatible large language model.
    """

    _model: Optional[BaseChatModel] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._credential = credential
        self._model_name = model_name
        self._model = ChatOpenAI(
            model=model_name,
            api_key=credential.get("api_key") or "empty",
            base_url=credential.get("endpoint_url", ""),
        )

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return self._model.invoke(input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return await self._model.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        self._model.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self._model.astream(input_dict, config=config, **kwargs):
            yield chunk

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        self._tools = tools
        if self._model is not None:
            self._model = self._model.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        def wrapper(input_data: LanguageModelInput, config: Optional[RunnableConfig] = None):
            structured_llm = self._model.with_structured_output(schema=schema, **kwargs)

            return structured_llm.invoke(input_data, config=config)

        return RunnableLambda(wrapper)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_base": self._credential.get("endpoint_url", ""),
            "openai_api_type": "openai",
            "openai_api_key": self._credentialget("api_key", ""),
            "openai_api_deployment": self._model_name,
        }


class OAIAPICompatEmbeddingModel(AIModel):

    _model: Optional[Embeddings] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = OAICompatEmbeddingModel(
            model_name=model_name,
            credential=credential,
        )
        self._credential = credential
        self._model_name = model_name

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self._model.embed_documents(texts)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self._model.aembed_documents(texts)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self._model.embed_query(text)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self._model.aembed_query(text)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_base": self._credential.get("endpoint_url", ""),
            "openai_api_type": "openai",
            "openai_api_key": self._credential.get("api_key", ""),
            "openai_api_deployment": self._model_name,
        }


class OAICompatRerankModel(AIModel):

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("api_key"):
                credential["api_key"] = aes_client.decrypt(credential["api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._credential = credential
        self._model_name = model_name

    def rerank(
        self,
        query: str,
        docs: list[str],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
    ) -> RerankResult:
        if len(docs) == 0:
            return RerankResult(docs=[])

        server_url = self._credential["endpoint_url"]
        model_name = self._model_name

        if not server_url:
            raise CredentialsValidateFailedError("server_url is required")
        if not model_name:
            raise CredentialsValidateFailedError("model_name is required")

        url = server_url
        headers = {
            "Content-Type": "application/json",
        }
        if self._credential.get("api_key"):
            headers["Authorization"] = f"Bearer {self._credential['api_key']}"

        # TODO: Do we need truncate docs to avoid llama.cpp return error?

        data = {
            "model": model_name,
            "query": query,
            "documents": docs,
            "top_n": top_n,
            "return_documents": True,
        }

        try:
            response = post(str(URL(url) / "rerank"), headers=headers, data=dumps(data), timeout=60)
            response.raise_for_status()
            results = response.json()

            rerank_documents = []
            scores = [result["relevance_score"] for result in results["results"]]

            # Min-Max Normalization: Normalize scores to 0 ~ 1.0 range
            min_score = min(scores)
            max_score = max(scores)
            score_range = max_score - min_score if max_score != min_score else 1.0  # Avoid division by zero

            for result in results["results"]:
                index = result["index"]

                # Retrieve document text (fallback if llama.cpp rerank doesn't return it)
                text = docs[index]
                document = result.get("document", {})
                if document:
                    if isinstance(document, dict):
                        text = document.get("text", docs[index])
                    elif isinstance(document, str):
                        text = document

                # Normalize the score
                normalized_score = (result["relevance_score"] - min_score) / score_range

                # Create RerankDocument object with normalized score
                rerank_document = RerankDocument(
                    index=index,
                    text=text,
                    score=normalized_score,
                )

                # Apply threshold (if defined)
                if score_threshold is None or normalized_score >= score_threshold:
                    rerank_documents.append(rerank_document)

            # Sort rerank_documents by normalized score in descending order
            rerank_documents.sort(key=lambda doc: doc.score, reverse=True)

            return RerankResult(docs=rerank_documents)

        except httpx.HTTPStatusError as e:
            raise InvokeServerUnavailableError(str(e))

    def unified_credential(self) -> Dict:
        return {
            "openai_api_base": self._credential.get("endpoint_url", ""),
            "openai_api_type": "openai",
            "openai_api_key": self._credential.get("api_key", ""),
            "openai_api_deployment": self._model_name,
        }


class OAICompatEmbeddingModel(Embeddings):
    """
    Model class for Openai compatible embedding model.
    """

    def __init__(self, model_name: str, credential: Dict):
        self._model_name = model_name
        self._credentials = credential

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embed search docs.

        Args:
            texts: List of text to embed.

        Returns:
            List of embeddings.
        """
        return self._invoke_documents(texts)

    def embed_query(self, text: str) -> list[float]:
        """Embed query text.

        Args:
            text: Text to embed.

        Returns:
            Embedding.
        """
        return self._invoke_query(text)

    async def aembed_documents(self, texts: list[str]) -> list[list[float]]:
        """Asynchronous Embed search docs.

        Args:
            texts: List of text to embed.

        Returns:
            List of embeddings.
        """
        return self._invoke_documents(texts)

    async def aembed_query(self, text: str) -> list[float]:
        """Asynchronous Embed query text.

        Args:
            text: Text to embed.

        Returns:
            Embedding.
        """
        return self._invoke_query(text)

    def _invoke_query(self, query: str) -> list[float]:
        headers = {"Content-Type": "application/json"}

        api_key = self._credentials.get("api_key")
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        endpoint_url = self._credentials.get("endpoint_url")
        assert endpoint_url is not None, "endpoint_url is required in credentials"
        if not endpoint_url.endswith("/"):
            endpoint_url += "/"

        endpoint_url = urljoin(endpoint_url, "embeddings")

        extra_model_kwargs = {}

        extra_model_kwargs["encoding_format"] = "float"

        payload = {"input": query, "model": self._model_name, **extra_model_kwargs}

        # Make the request to the OpenAI API
        response = requests.post(endpoint_url, headers=headers, data=json.dumps(payload), timeout=(10, 300))

        response.raise_for_status()  # Raise an exception for HTTP errors
        response_data = response.json()

        # Extract embeddings and used tokens from the response
        embeddings_batch = [data["embedding"] for data in response_data["data"]]
        if len(embeddings_batch) != 1:
            raise ValueError("embeddings_batch length less than 1")

        return embeddings_batch[0]

    def _invoke_documents(self, texts: list[str]) -> list[list[float]]:
        """
        Invoke text embedding model

        :param model: model name
        :param credentials: model credentials
        :param texts: texts to embed
        :param user: unique user id
        :param input_type: input type
        :return: embeddings result
        """

        # Prepare headers and payload for the request
        headers = {"Content-Type": "application/json"}

        api_key = self._credentials.get("api_key")
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        endpoint_url = self._credentials.get("endpoint_url")
        assert endpoint_url is not None, "endpoint_url is required in credentials"
        if not endpoint_url.endswith("/"):
            endpoint_url += "/"

        endpoint_url = urljoin(endpoint_url, "embeddings")

        extra_model_kwargs = {}
        extra_model_kwargs["encoding_format"] = "float"

        # get model properties
        context_size: int = int(self._credentials.get("context_size", 4096))
        max_chunks: int = self._credentials.get("max_chunks", 1000)

        inputs = []
        indices = []
        used_tokens = 0

        for i, text in enumerate(texts):
            # Here token count is only an approximation based on the GPT2 tokenizer
            # TODO: Optimize for better token estimation and chunking
            num_tokens = self._get_num_tokens_by_gpt2(text)

            if num_tokens >= context_size:
                cutoff = int(np.floor(len(text) * (context_size / num_tokens)))
                # if num tokens is larger than context length, only use the start
                inputs.append(text[0:cutoff])
            else:
                inputs.append(text)
            indices += [i]

        batched_embeddings = []
        _iter = range(0, len(inputs), max_chunks)

        for i in _iter:
            # Prepare the payload for the request
            payload = {"input": inputs[i : i + max_chunks], "model": self._model_name, **extra_model_kwargs}

            # Make the request to the OpenAI API
            response = requests.post(
                endpoint_url, headers=headers, data=json.dumps(payload), timeout=(10, 300)
            )

            response.raise_for_status()  # Raise an exception for HTTP errors
            response_data = response.json()

            # Extract embeddings and used tokens from the response
            embeddings_batch = [data["embedding"] for data in response_data["data"]]
            embedding_used_tokens = response_data["usage"]["total_tokens"]

            used_tokens += embedding_used_tokens
            batched_embeddings += embeddings_batch

        return batched_embeddings

    def _get_num_tokens_by_gpt2(self, text: str) -> int:
        """
        use gpt2 tokenizer to get num tokens
        """
        _tokenizer = OAICompatEmbeddingModel.get_encoder()
        tokens = _tokenizer.encode(text)
        return len(tokens)

    @staticmethod
    def get_encoder() -> Any:
        global _tokenizer, _lock
        with _lock:
            if _tokenizer is None:
                # Try to use tiktoken to get the tokenizer because it is faster
                #
                try:
                    import tiktoken

                    _tokenizer = tiktoken.get_encoding("gpt2")
                except Exception:
                    from os.path import abspath, dirname, join

                    from transformers import (
                        GPT2Tokenizer as TransformerGPT2Tokenizer,  # type: ignore
                    )

                    base_path = abspath(__file__)
                    gpt2_tokenizer_path = join(dirname(base_path), "gpt2")
                    _tokenizer = TransformerGPT2Tokenizer.from_pretrained(gpt2_tokenizer_path)

            return _tokenizer



from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Type, Union

import dashscope
from langchain_community.chat_models.tongyi import ChatTongyi
from langchain_community.embeddings.dashscope import DashScopeEmbeddings
from langchain_core.embeddings.embeddings import Embeddings
from langchain_core.language_models.base import LanguageModelInput
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages.base import BaseMessage, BaseMessageChunk
from langchain_core.runnables.base import Runnable, RunnableLambda
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools.base import BaseTool
from pydantic.main import BaseModel

from llm_sdk.model_providers.aes import AESCipher
from llm_sdk.model_providers.base import AIModel
from llm_sdk.model_providers.rerank import RerankDocument, RerankModel, RerankResult

import logging

logger = logging.getLogger(__name__)


class TongyiLargeLanguageModel(AIModel):
    _model: Optional[BaseChatModel] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("dashscope_api_key"):
                credential["dashscope_api_key"] = aes_client.decrypt(credential["dashscope_api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = ChatTongyi(
            model=model_name,
            api_key=credential.get("dashscope_api_key", ""),
            max_retries=credential.get("max_retries", 3),
        )
        self._model_name = model_name
        self._credential = credential

    def invoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return self._model.invoke(input_dict, config=config, **kwargs)

    async def ainvoke(
        self,
        input_dict: Union[List[BaseMessage], List[str], str],
        config: Optional[RunnableConfig] = None,
        **kwargs,
    ) -> Any:
        return await self._model.ainvoke(input_dict, config=config, **kwargs)

    def stream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> Iterator[BaseMessageChunk]:
        self._model.stream(input_dict, config=config, **kwargs)

    async def astream(
        self, input_dict: List[BaseMessage], config: Optional[RunnableConfig] = None, **kwargs
    ) -> AsyncGenerator[BaseMessageChunk, None]:
        async for chunk in self._model.astream(input_dict, config=config, **kwargs):
            yield chunk

    def bind_tools(self, tools: List[BaseTool], **kwargs) -> "AIModel":
        self._tools = tools
        if self._model is not None:
            self._model = self._model.bind_tools(tools, **kwargs)
        return self

    def with_structured_output(self, schema: Union[Dict, Type[BaseModel]], **kwargs) -> Runnable:
        def wrapper(input_data: LanguageModelInput, config: Optional[RunnableConfig] = None):
            structured_llm = self._model.with_structured_output(schema=schema, **kwargs)

            return structured_llm.invoke(input_data, config=config)

        return RunnableLambda(wrapper)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("dashscope_api_key", ""),
            "openai_api_type": "tongyi",
            "openai_api_deployment": self._model_name,
            "timeout": self._credential.get("timeout", 60),
            "max_retries": self._credential.get("max_retries", 1),
            "temperature": self._credential.get("temperature", 0.7),
        }


class TongyiTextEmbeddingModel(AIModel):

    _model: Optional[Embeddings] = None

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("dashscope_api_key"):
                credential["dashscope_api_key"] = aes_client.decrypt(credential["dashscope_api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._model = DashScopeEmbeddings(
            dashscope_api_key=credential.get("dashscope_api_key", ""),
            model=model_name,
            max_retries=credential.get("max_retries", 3),
        )
        self._credential = credential
        self._model_name = model_name

    def embed_documents(self, texts: List[str], config: Optional[RunnableConfig] = None) -> List[List[float]]:
        return self._model.embed_documents(texts)

    async def aembed_documents(
        self, texts: List[str], config: Optional[RunnableConfig] = None
    ) -> List[List[float]]:
        return await self._model.aembed_documents(texts)

    def embed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return self._model.embed_query(text)

    async def aembed_query(self, text: str, config: Optional[RunnableConfig] = None) -> List[float]:
        return await self._model.aembed_query(text)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("dashscope_api_key", ""),
            "openai_api_type": "tongyi",
            "openai_api_deployment": self._model_name,
        }


class GTERerankModel(RerankModel):

    def __init__(self, model_name: str, credential: Dict, app_id: Optional[str] = None):
        super().__init__()

        if app_id is None:
            raise ValueError("app_id is required")
        try:
            aes_client = AESCipher(app_id)
            # decrypt credential
            if credential.get("dashscope_api_key"):
                credential["dashscope_api_key"] = aes_client.decrypt(credential["dashscope_api_key"])
        except Exception as e:
            logger.warning(f"decrypt credential failed: {e}, use original key")

        self._credential = credential
        self._model_name = model_name

    def rerank(
        self,
        query: str,
        docs: list[str],
        score_threshold: Optional[float] = None,
        top_n: Optional[int] = None,
    ) -> RerankResult:
        if len(docs) == 0:
            return RerankResult(docs=[])

        # initialize client
        dashscope.api_key = self._credential.get("dashscope_api_key", "")

        response = dashscope.TextReRank.call(
            query=query,
            documents=docs,
            model=self._model_name,
            top_n=top_n,
            return_documents=True,
        )

        rerank_documents: list[RerankDocument] = []
        if not response.output:
            return RerankResult(docs=rerank_documents)
        for _, result in enumerate(response.output.results):
            # format document
            rerank_document = RerankDocument(
                index=result.index,
                score=result.relevance_score,
                text=result["document"]["text"],
            )

            # score threshold check
            if score_threshold is not None:
                if result.relevance_score >= score_threshold:
                    rerank_documents.append(rerank_document)
            else:
                rerank_documents.append(rerank_document)

        return RerankResult(docs=rerank_documents)

    def unified_credential(self) -> Dict:
        return {
            "openai_api_key": self._credential.get("dashscope_api_key", ""),
            "openai_api_type": "tongyi",
            "openai_api_deployment": self._model_name,
        }
